from enum import Enum
from pydantic import Field, model_validator
from jsonschema_rs import validator_for
import json
from importlib.resources import files
from jsonschema_rs import ValidationError


from enum import Enum

from visivo.models.trace_props.json_schema_base import JsonSchemaBase, get_message_from_error


class TraceType(str, Enum):
    BAR = "bar"
    BARPOLAR = "barpolar"
    BOX = "box"
    CANDLESTICK = "candlestick"
    CARPET = "carpet"
    CHOROPLETH = "choropleth"
    CHOROPLETHMAP = "choroplethmap"
    CHOROPLETHMAPBOX = "choroplethmapbox"
    CONE = "cone"
    CONTOUR = "contour"
    CONTOURCARPET = "contourcarpet"
    DENSITYMAP = "densitymap"
    DENSITYMAPBOX = "densitymapbox"
    FUNNEL = "funnel"
    FUNNELAREA = "funnelarea"
    HEATMAP = "heatmap"
    HISTOGRAM = "histogram"
    HISTOGRAM2D = "histogram2d"
    HISTOGRAM2DCONTOUR = "histogram2dcontour"
    ICICLE = "icicle"
    IMAGE = "image"
    INDICATOR = "indicator"
    ISOSURFACE = "isosurface"
    MESH3D = "mesh3d"
    OHLC = "ohlc"
    PARCATS = "parcats"
    PARCOORDS = "parcoords"
    PIE = "pie"
    SANKEY = "sankey"
    SCATTER = "scatter"
    SCATTER3D = "scatter3d"
    SCATTERCARPET = "scattercarpet"
    SCATTERGEO = "scattergeo"
    SCATTERGL = "scattergl"
    SCATTERMAP = "scattermap"
    SCATTERMAPBOX = "scattermapbox"
    SCATTERPOLAR = "scatterpolar"
    SCATTERPOLARGL = "scatterpolargl"
    SCATTERSMITH = "scattersmith"
    SCATTERTERNARY = "scatterternary"
    SPLOM = "splom"
    STREAMTUBE = "streamtube"
    SUNBURST = "sunburst"
    SURFACE = "surface"
    TREEMAP = "treemap"
    VIOLIN = "violin"
    VOLUME = "volume"
    WATERFALL = "waterfall"


class TraceProps(JsonSchemaBase):

    type: TraceType = Field(..., description="Type of the trace")

    @model_validator(mode="after")
    def validate_against_schema(self) -> "TraceProps":
        if not hasattr(TraceProps, "_validators"):
            TraceProps._validators = {}
        if not hasattr(TraceProps, "_schemas"):
            TraceProps._schemas = {}

        if self.type.value not in TraceProps._validators:
            schema_path = files("visivo.schema").joinpath(f"{self.type.value}.schema.json")
            with open(schema_path) as f:
                schema = json.load(f)
            TraceProps._schemas[self.type.value] = schema
            TraceProps._validators[self.type.value] = validator_for(schema)

        validator = TraceProps._validators.get(self.type.value)
        if not validator:
            raise ValueError(f"Schema not found for trace type: {self.type.value}")

        try:
            data_dict = self.model_dump()

            validator.validate(data_dict)

        except FileNotFoundError:
            raise ValueError(f"Schema file not found for trace type: {self.type.value}")
        except json.JSONDecodeError:
            raise ValueError(f"Invalid JSON in schema file for trace type: {self.type.value}")
        except ValidationError as e:
            schema = TraceProps._schemas.get(self.type.value)
            message = get_message_from_error(e, schema)
            raise ValueError(
                f"Validation error for trace type {self.type.value} at location: {e.instance_path}: {str(message)}"
            )

        return self



from visivo.server.models.base import (
    Base,
    Column,
    String,
    Integer,
    Boolean,
    ForeignKey,
    relationship,
)


class SessionStateModel(Base):
    __tablename__ = "session_state"

    worksheet_id = Column(String, ForeignKey("worksheets.id", ondelete="CASCADE"), primary_key=True)
    tab_order = Column(Integer, nullable=False)
    is_visible = Column(Boolean, default=True)

    worksheet = relationship("WorksheetModel", back_populates="session_state")

    def to_dict(self):
        return {
            "worksheet_id": self.worksheet_id,
            "tab_order": self.tab_order,
            "is_visible": self.is_visible,
        }



import PyInstaller.__main__
from pathlib import Path

HERE = Path(__file__).parent.absolute()
path_to_main = str(HERE / "command_line.py")


def build():
    PyInstaller.__main__.run(
        [
            path_to_main,
            "--clean",
            "--onedir",
            "--noconfirm",
            "--collect-submodules",
            "engineio",
            "--collect-submodules",
            "socketio",
            "--collect-submodules",
            "flask_socketio",
            "-n",
            "visivo",
            "--add-data",
            "visivo/schema/*.json:visivo/schema",
            "--add-data",
            "visivo/templates/queries/*:visivo/templates/queries",
            "--add-data",
            "visivo/templates/charts/*:visivo/templates/charts",
            "--add-data",
            "visivo/viewers/*:visivo/viewers",
        ]
    )



from typing import Literal, Optional
from visivo.models.sources.sqlalchemy_source import SqlalchemySource
from pydantic import Field, SecretStr
import os

BigQueryType = Literal["bigquery"]


class BigQuerySource(SqlalchemySource):
    """
    BigQuerySources hold the connection information to Google BigQuery instances.

    !!! info "BigQuery Authentication"
        You can authenticate BigQuery in one of two ways:
         1. Pass a base64 encoded service account key to the `credentials_base64` field.
         2. Set the absolute file path to the credentials file in a environment variable named `GOOGLE_APPLICATION_CREDENTIALS`.

        === "Base64 Encoded Credentials"
            Using encoding the service key json credential file to base64 can be a useful way to authenticate BigQuery
            without logging into Google Cloud Console each time and makes it easier to manage credentials in CI/CD pipelines.

            However utilizing base64 encoding requires a few extra steps:

            1. Create a Google Cloud Service Account
                1. Go to the [Google Cloud Console](https://console.cloud.google.com)
                2. Select your project
                3. Navigate to "IAM & Admin" > "Service Accounts"
                4. Click "Create Service Account"
                5. Give it a name and description
                6. Grant it the "BigQuery Admin" role (or more restrictive custom role)
                7. Click "Done"
            2. Create and download credentials
                1. Find your service account in the list
                2. Click the three dots menu > "Manage keys"
                3. Click "Add Key" > "Create new key"
                4. Choose JSON format
                5. Click "Create" - this downloads your credentials file
            3. Convert credentials to base64
                ```bash
                # On Linux/Mac
                python -m base64 < credentials.json > encoded.txt

                # On Windows PowerShell
                [Convert]::ToBase64String([System.IO.File]::ReadAllBytes("credentials.json")) > encoded.txt
                ```
            4. Use the contents of encoded.txt as your credentials_base64 value. You can store the single line key in your untracked env file and use the `{% raw %}{{ env_var('VAR_NAME') }}{% endraw %}` syntax to reference the environment variable in your Visivo config.

        === "`GOOGLE_APPLICATION_CREDENTIALS` Environment Variable"
            If you use gcloud locally you probably have this environment variable configured already.

            Run `echo $GOOGLE_APPLICATION_CREDENTIALS` in your terminal. If it returns your crendetials then
            you're all set. and can configure a BigQuerySource without the `credentials_base64` field.

            If you don't have the environment variable, follow these steps:

            1. Create a Google Cloud Service Account
                1. Go to the [Google Cloud Console](https://console.cloud.google.com)
                2. Select your project
                3. Navigate to "IAM & Admin" > "Service Accounts"
                4. Click "Create Service Account"
                5. Give it a name and description
                6. Grant it the "BigQuery Admin" role (or more restrictive custom role)
                7. Click "Done"
            2. Create and download credentials
                1. Find your service account in the list
                2. Click the three dots menu > "Manage keys"
                3. Click "Add Key" > "Create new key"
                4. Choose JSON format
                5. Click "Create" - this downloads your credentials file
            3. Set the environment variable
                You can set the environment variable in your shell profile file.
                ```bash
                export GOOGLE_APPLICATION_CREDENTIALS="/path/to/your/credentials.json"
                ```
                or in your untracked .env file.
                ```
                GOOGLE_APPLICATION_CREDENTIALS=/path/to/your/credentials.json
                ```
            This method is easier to manage and does not require any extra steps to authenticate.

        The service account needs at minimum the "BigQuery User" role to execute queries.
        For more restricted access, you can create a custom role with just the required permissions:

        - bigquery.jobs.create
        - bigquery.tables.get
        - bigquery.tables.getData
        - bigquery.tables.list

    !!! example

        === "Simple"

            ``` yaml
                sources:
                  - name: bigquery_source
                    type: bigquery
                    project: my-project-id
                    database: my_dataset
                    credentials_base64: {% raw %}{{ env_var('BIGQUERY_BASE64_ENCODED_CREDENTIALS') }}{% endraw %}
            ```

    Note: Recommended environment variable use is covered in the [sources overview.](/topics/sources/)
    """

    project: str = Field(
        description="The Google Cloud project ID that contains your BigQuery dataset."
    )

    credentials_base64: Optional[SecretStr] = Field(
        None,
        description="The Google Cloud service account credentials JSON string base64 encoded. Turn your JSON into a base64 string in the command line with `python -m base64 < credentials.json > encoded.txt`. Not required if GOOGLE_APPLICATION_CREDENTIALS environment variable is set. ",
    )

    database: Optional[str] = Field(
        None,
        description="The default BigQuery dataset to use for queries.",
    )

    type: BigQueryType
    connection_pool_size: Optional[int] = Field(
        8, description="The pool size that is used for this connection."
    )

    def get_dialect(self):
        return "bigquery"

    def url(self):
        base_url = f"bigquery://{self.project}"
        if self.database:
            base_url += f"/{self.database}"

        credentials = (
            self.credentials_base64.get_secret_value() if self.credentials_base64 else None
        )

        # Check for either credentials_base64 or GOOGLE_APPLICATION_CREDENTIALS
        if not credentials and not os.getenv("GOOGLE_APPLICATION_CREDENTIALS"):
            raise ValueError(
                "Authentication credentials not found. Either provide credentials_base64 or set GOOGLE_APPLICATION_CREDENTIALS environment variable."
            )

        # Only append credentials to URL if using base64 method
        if credentials:
            base_url += f"?credentials_base64={credentials}"

        return base_url



from typing import List, Optional
from visivo.models.alert import Alert
from visivo.models.dag import all_descendants_of_type

from visivo.models.destinations.fields import DestinationField
from visivo.models.include import Include
from visivo.models.models.model import Model
from visivo.models.models.fields import ModelField
from visivo.models.models.sql_model import SqlModel
from visivo.models.selector import Selector, SelectorType
from visivo.models.sources.fields import SourceField

from visivo.models.base.parent_model import ParentModel
from visivo.models.base.base_model import REF_REGEX
from visivo.models.base.context_string import INLINE_REF_REGEX
from visivo.models.dashboards.fields import DashboardField
from visivo.models.chart import Chart
from visivo.models.trace import Trace
from visivo.models.table import Table
from visivo.models.defaults import Defaults
from visivo.models.dbt import Dbt
from typing import List
from visivo.models.base.named_model import NamedModel
from visivo.models.base.base_model import BaseModel
from pydantic import ConfigDict, Field, model_validator
from visivo.utils import PROJECT_CHILDREN
from click import ClickException
from visivo.version import VISIVO_VERSION


class Project(NamedModel, ParentModel):
    model_config = ConfigDict(populate_by_name=True, extra="forbid")

    defaults: Optional[Defaults] = None
    dbt: Optional[Dbt] = None
    project_file_path: Optional[str] = None
    cli_version: Optional[str] = Field(
        default=VISIVO_VERSION,
        description="The version of the CLI that created the project.",
    )
    includes: List[Include] = []
    destinations: List[DestinationField] = []
    alerts: List[Alert] = []
    sources: List[SourceField] = Field(
        [],
        description="A list of source objects.",
        alias="targets",
    )
    models: List[ModelField] = []
    traces: List[Trace] = []
    tables: List[Table] = []
    charts: List[Chart] = []
    selectors: List[Selector] = []
    dashboards: List[DashboardField] = []

    def child_items(self) -> List:
        project_children = PROJECT_CHILDREN.copy()
        children = []
        for child_type in project_children:
            items = getattr(self, child_type, [])
            children.extend(items)
        return children

    def named_child_nodes(self) -> dict:
        """
        Returns a dictionary of all named child nodes of the project, independent of the
        literal position of the child in the project file. This enables us to find
        all children even if they are nested in a dashboard or chart.
        """
        dag = self.dag()
        named_nodes = {}
        # First pass - collect all nodes and their inline dependencies
        for node in dag.nodes():
            if hasattr(node, "name"):
                if node.name is not None:
                    is_named = True
                else:
                    is_named = False
            else:
                is_named = False
            if is_named and (Project.is_project_child(node) or isinstance(node, Project)):
                fully_referenced_model_dump = Project._fully_referenced_model_dump(node)
                file_path = fully_referenced_model_dump.pop("file_path", self.project_file_path)
                path = fully_referenced_model_dump.pop("path", "Not Found")
                inline_dependent_objects = fully_referenced_model_dump.pop(
                    "inline_dependent_objects", []
                )
                _ = fully_referenced_model_dump.pop("changed", "Not Found")
                direct_children = dag.get_named_children(node.name)
                direct_parents = dag.get_named_parents(node.name)
                contents = {
                    "type": node.__class__.__name__,
                    "type_key": Project.get_key_for_project_child_class(node.__class__.__name__),
                    "config": fully_referenced_model_dump,
                    "file_path": file_path,
                    "new_file_path": file_path,
                    "path": path,
                    "inline_dependent_objects": inline_dependent_objects,
                    "inline_object_parents": [],  # Initialize empty list for parents
                    "is_inline_defined": False,
                    "direct_children": direct_children,
                    "direct_parents": direct_parents,
                }
                named_nodes[node.name] = contents

        # Second pass - build the reverse mapping
        for node_name, node_data in named_nodes.items():
            if node_data["type"] != "Project":
                for dependent_object in node_data["inline_dependent_objects"]:
                    if dependent_object in named_nodes:
                        named_nodes[dependent_object]["inline_object_parents"].append(node_name)
                        named_nodes[dependent_object]["is_inline_defined"] = True
        return named_nodes

    @classmethod
    def _fully_referenced_model_dump(cls, node: ParentModel) -> dict:
        import json
        import re

        def clean_value(value, inline_dependent_objects: list):
            # Case 1: Value is a dictionary
            if isinstance(value, dict):
                # Check if it's a project child and has a non-null 'name'
                if (
                    Project.is_type_project_child(value.get("__type__", "na"))
                    and value.get("name") is not None
                ):
                    inline_defined_named_child = json.dumps(
                        {"name": value["name"], "is_inline_defined": True}
                    )
                    inline_dependent_objects.append(value["name"])
                    return inline_defined_named_child
                # If not replaced, recurse into the dictionary's values
                return {k: clean_value(v, inline_dependent_objects) for k, v in value.items()}
            # Case 2: Value is a list
            elif isinstance(value, list):
                return [clean_value(elem, inline_dependent_objects) for elem in value]
            # Case 3: Value is a primitive (str, int, float, bool, None)
            elif isinstance(value, str):
                # Check for inline references ${ref(Name)}
                inline_matches = re.search(INLINE_REF_REGEX, value)
                if inline_matches and value.strip() == inline_matches.group(0):
                    ref_name = inline_matches.group(1).strip()
                    return json.dumps(
                        {"name": ref_name, "is_inline_defined": False, "original_value": value}
                    )

                # Check for direct ref(Name) pattern
                direct_matches = re.search(REF_REGEX, value)
                if direct_matches:
                    ref_name = direct_matches.group("ref_name").strip()
                    return json.dumps(
                        {"name": ref_name, "is_inline_defined": False, "original_value": value}
                    )
                return value
            else:
                return value

        def remove_type_keys(data):
            if isinstance(data, dict):
                return {k: remove_type_keys(v) for k, v in data.items() if k != "__type__"}
            elif isinstance(data, list):
                return [remove_type_keys(elem) for elem in data]
            else:
                return data

        model_dump_json_string = node.model_dump_json(
            exclude_none=True, context={"include_type": True}
        )
        jsonable_model_dump = json.loads(model_dump_json_string)
        inline_dependent_objects = []
        fully_referenced_project_child_dict = {
            k: clean_value(v, inline_dependent_objects) for k, v in jsonable_model_dump.items()
        }
        fully_referenced_project_child_dict = remove_type_keys(fully_referenced_project_child_dict)
        fully_referenced_project_child_dict["inline_dependent_objects"] = inline_dependent_objects
        return fully_referenced_project_child_dict

    @model_validator(mode="after")
    def validate_cli_version(self):
        if self.cli_version != VISIVO_VERSION:
            raise ClickException(
                f"The project specifies {self.cli_version}, but the current version of visivo installed is {VISIVO_VERSION}. Your project version needs to match your CLI version."
            )
        return self

    @model_validator(mode="after")
    def validate_default_names(self):
        sources, alerts = (self.sources, self.alerts)
        source_names = [source.name for source in sources]
        alert_names = [alert.name for alert in alerts]
        defaults = self.defaults
        if not defaults:
            return self

        if defaults.source_name and defaults.source_name not in source_names:
            raise ValueError(f"default source '{defaults.source_name}' does not exist")

        if defaults.alert_name and defaults.alert_name not in alert_names:
            raise ValueError(f"default alert '{defaults.alert_name}' does not exist")

        return self

    @model_validator(mode="after")
    def validate_models_have_sources(self):
        defaults = self.defaults
        if defaults and defaults.source_name:
            return self

        for model in self.descendants_of_type(Model):
            if isinstance(model, SqlModel) and not model.source:
                raise ValueError(
                    f"'{model.name}' does not specify a source and project does not specify default source"
                )

        return self

    @model_validator(mode="after")
    def validate_dag(self):

        dag = self.dag()
        if not dag.validate_dag():
            raise ValueError("Project contains a circular reference.")

        tables = all_descendants_of_type(type=Table, dag=dag, from_node=self)
        for table in tables:
            selectors = all_descendants_of_type(type=Selector, dag=dag, from_node=table)
            if len(selectors) > 0 and selectors[0].type == SelectorType.multiple:
                raise ValueError(
                    f"Table with name '{table.name}' has a selector with a 'multiple' type.  This is not permitted."
                )

        return self

    @model_validator(mode="after")
    def validate_project_is_sole_root_node(self):
        dag = self.dag()
        roots = dag.get_root_nodes()
        if len(roots) > 1:
            root_list = ", ".join([root.__class__.__name__ for root in roots])
            raise ValueError(
                f"Project must be the sole root node in the DAG. Current root nodes: {root_list}"
            )
        elif len(roots) == 0:
            raise ValueError("No root nodes found in the DAG. Please add a name for your project.")
        elif len(roots) == 1:
            root = roots[0]
            if root.__class__.__name__ != "Project":
                raise ValueError("The sole root node in the DAG must be a Project.")
        return self

    @model_validator(mode="after")
    def validate_names(self):
        Project.traverse_names([], self)
        return self

    @model_validator(mode="before")
    def set_path_on_named_models(cls, values):
        def set_path_recursively(obj, path=""):
            if isinstance(obj, dict):
                obj["path"] = path
                for key, value in obj.items():
                    if key not in ["props", "defaults", "layout", "columns"]:
                        new_path = f"{path}.{key}" if path else key
                        set_path_recursively(value, new_path)
            elif isinstance(obj, list):
                for index, item in enumerate(obj):
                    new_path = f"{path}[{index}]"
                    set_path_recursively(item, new_path)

        set_path_recursively(values, "project")
        return values

    @classmethod
    def get_child_objects(cls) -> dict:
        """
        Returns a dictionary mapping each project child type to its field object.
        This is used to identify which objects in the project hierarchy are direct children of the project.
        """
        child_objects = {}
        for field_name in PROJECT_CHILDREN:
            field = cls.model_fields[field_name]
            child_objects[field_name] = field
        return child_objects

    @classmethod
    def is_type_project_child(cls, object_type: str) -> bool:
        """
        Accepts a string and returns True if it is a child of the project.
        """
        from re import search

        project_child_objects = cls.get_child_objects()

        search_str = rf"[\s\[]{object_type}[,\]]"
        is_match = search(search_str, str(project_child_objects)) is not None

        return is_match

    @classmethod
    def is_project_child(cls, obj) -> list:
        """
        Accepts an object and returns True if it is a child of the project.
        """
        from re import search

        project_child_objects = cls.get_child_objects()
        if isinstance(obj, BaseModel):
            obj_name = obj.__class__.__name__
            search_str = rf"[\s\[]{obj_name}[,\]]"
            is_match = search(search_str, str(project_child_objects)) is not None
            return is_match
        return False

    @classmethod
    def get_key_for_project_child_class(cls, project_child_class: str) -> str:
        """
        Accepts an object and returns True if it is a child of the project.
        """
        from re import search

        project_child_objects = cls.get_child_objects()
        for key, value in project_child_objects.items():
            search_str = rf"[\s\[]{project_child_class}[,\]]"
            is_match = search(search_str, str(value)) is not None
            if is_match:
                return key
        if project_child_class == "Project":
            return "na"
        raise ValueError(f"Project child class '{project_child_class}' not found in project")

    @classmethod
    def traverse_names(cls, names, object):
        if isinstance(object, ParentModel):
            for child_item in object.child_items():
                if isinstance(child_item, BaseModel) and hasattr(child_item, "name"):
                    name = NamedModel.get_name(obj=child_item)
                    if name in names:
                        raise ValueError(
                            f"{child_item.__class__.__name__} name '{name}' is not unique in the project"
                        )
                    if name:
                        names.append(name)
                Project.traverse_names(names, child_item)



from typing import get_args
import click
import yaml
import json
import os
from pathlib import Path
from visivo.logger.logger import Logger
from visivo.models.include import Include
from visivo.models.models.sql_model import SqlModel
from visivo.models.sources.postgresql_source import PostgresqlSource, PostgresqlType
from visivo.models.sources.snowflake_source import SnowflakeSource, SnowflakeType
from visivo.models.sources.sqlite_source import SqliteSource, SqliteType
from visivo.models.sources.mysql_source import MysqlSource, MysqlType
from visivo.models.sources.bigquery_source import BigQuerySource, BigQueryType
from visivo.models.project import Project
from visivo.models.dashboard import Dashboard
from visivo.models.chart import Chart
from visivo.models.item import Item
from visivo.models.row import Row
from visivo.models.sources.sqlite_source import SqliteSource
from visivo.models.trace import Trace
from visivo.models.defaults import Defaults
from visivo.commands.utils import create_file_database
from visivo.parsers.file_names import PROFILE_FILE_NAME
from visivo.commands.utils import get_source_types
from visivo.models.sources.duckdb_source import DuckdbSource, DuckdbType
from visivo.version import VISIVO_VERSION


def init_phase(project_dir):
    """Enables a quick set up by writing your source & api credentials to an env file."""
    user_home = os.path.expanduser("~")

    if project_dir:
        project_name = project_dir
    else:
        project_dir = "."
        project_name = os.path.basename(os.path.abspath("."))

    Logger.instance().success(f"Initializing in {project_dir}")
    sqlite_type = get_args(SqliteType)[0]
    postgresql_type = get_args(PostgresqlType)[0]
    mysql_type = get_args(MysqlType)[0]
    snowflake_type = get_args(SnowflakeType)[0]
    bigquery_type = get_args(BigQueryType)[0]
    duckdb_type = get_args(DuckdbType)[0]
    types = get_source_types()

    source_type = click.prompt("? Database type", type=click.Choice(types))
    if source_type == sqlite_type:
        source = SqliteSource(
            name="Example Source",
            database=f"{project_dir}/local.db",
            type=source_type,
        )
        create_file_database(source.url(), project_dir)
        source.database = "local.db"
        fp = open(f"{project_dir}/.env", "w+")
        fp.write("DB_PASSWORD=EXAMPLE_password_l0cation")
        fp.close()

    if source_type == postgresql_type:
        host = click.prompt("? Database host", type=str)
        database = click.prompt("? Database name", type=str)
        username = click.prompt("? Database username", type=str)
        password = click.prompt(
            "? Database password", type=str, hide_input=True, confirmation_prompt=True
        )
        fp = open(f"{project_dir}/.env", "w+")
        fp.write(f"DB_PASSWORD={password}")
        fp.close()
        source = PostgresqlSource(
            name="Example Source",
            host=host,
            database=database,
            type=source_type,
            password=password,
            username=username,
        )
    if source_type == mysql_type:
        host = click.prompt("? Database host", type=str)
        database = click.prompt("? Database name", type=str)
        username = click.prompt("? Database username", type=str)
        password = click.prompt(
            "? Database password", type=str, hide_input=True, confirmation_prompt=True
        )
        fp = open(f"{project_dir}/.env", "w+")
        fp.write(f"DB_PASSWORD={password}")
        fp.close()
        source = MysqlSource(
            name="Example Source",
            host=host,
            database=database,
            type=source_type,
            password=password,
            username=username,
        )

    if source_type == snowflake_type:
        database = click.prompt("? Database name", type=str)
        account = click.prompt("? Snowflake account", type=str)
        warehouse = click.prompt("? Snowflake warehouse", type=str)
        username = click.prompt("? Database username", type=str)
        password = click.prompt(
            "? Database password", type=str, hide_input=True, confirmation_prompt=True
        )
        fp = open(f"{project_dir}/.env", "w+")
        fp.write(f"DB_PASSWORD={password}")
        fp.close()
        source = SnowflakeSource(
            name="Example Source",
            database=database,
            type=source_type,
            password=password,
            username=username,
            account=account,
            warehouse=warehouse,
        )
    if source_type == bigquery_type:
        project = click.prompt("? BigQuery project", type=str)
        dataset = click.prompt("? BigQuery dataset", type=str)
        credentials_base64 = click.prompt(
            "? base64 encoded credentials",
            type=str,
            hide_input=True,
            confirmation_prompt=True,
        )
        fp = open(f"{project_dir}/.env", "w+")
        fp.write(f"DB_PASSWORD={credentials_base64}")
        fp.close()
        source = BigQuerySource(
            name="Example Source",
            project=project,
            database=dataset,
            type=source_type,
            credentials_base64=credentials_base64,
        )
    if source_type == duckdb_type:
        database = click.prompt("? Database file path", type=str, default=f"{project_dir}/local.db")
        source = DuckdbSource(
            name="Example Source",
            database=database,
            type=source_type,
        )
        create_file_database(source.url(), project_dir)
        source.database = "local.db"
        fp = open(f"{project_dir}/.env", "w+")
        fp.write("DB_PASSWORD=EXAMPLE_password_l0cation")
        fp.close()
    Logger.instance().debug("Generating project, gitignore & env files")

    model = SqlModel(name="Example Model", sql="select * from test_table")
    props = {"type": "scatter", "x": "?{x}", "y": "?{y}"}
    trace = Trace(name="Example Trace", model=model, props=props)
    chart = Chart(name="Example Chart", traces=[trace])
    item = Item(chart=chart)
    row = Row(items=[item])
    dashboard = Dashboard(name="Example Dashboard", rows=[row])
    defaults = Defaults(source_name=source.name)

    current_version = VISIVO_VERSION
    includes = Include(
        path=f"visivo-io/visivo.git@v{current_version} -- test-projects/demo/dashboards/welcome.visivo.yml"
    )
    project = Project(
        name=project_name,
        includes=[includes],
        defaults=defaults,
        sources=[source],
        dashboards=[dashboard],
    )

    fp = open(f"{project_dir}/project.visivo.yml", "w")
    fp.write(
        yaml.dump(json.loads(project.model_dump_json(exclude_none=True)), sort_keys=False).replace(
            "'**********'", "\"{{ env_var('DB_PASSWORD') }}\""
        )
    )
    fp.close()

    fp = open(f"{project_dir}/.gitignore", "w")
    fp.write(".env\ntarget\n.visivo_cache")
    fp.close()

    Logger.instance().success("Generated project, gitignore & env files")

    profile_path = f"{user_home}/.visivo/{PROFILE_FILE_NAME}"
    if not os.path.exists(profile_path):
        Logger.instance().info(
            f"> Visit 'https://app.visivo.io/profile' and create a new token if you don't already have one."
        )
        Logger.instance().info(
            f"> You may need to register or get added to an account before visiting your profile."
        )
        token = click.prompt("? Personal token", type=str)
        os.makedirs(f"{user_home}/.visivo", exist_ok=True)
        fp = open(profile_path, "w")
        fp.write(f"token: {token}")
        fp.close()
        Logger.instance().info(f"> Created profile in '~/.visivo/profile.yml'")
    else:
        message = "Found profile at location: " + profile_path
        Logger.instance().info(message)
    Logger.instance().info(f"> Created project in '{project_name}'")



from visivo.logger.logger import Logger
from visivo.models.sources.source import Source
from visivo.jobs.job import (
    Job,
    JobResult,
    format_message_failure,
    format_message_success,
    start_message,
)
from time import time


def action(source_to_test: Source):
    Logger.instance().info(start_message("Source", source_to_test))
    try:
        start_time = time()
        source_to_test.read_sql("select 1")
        success_message = format_message_success(
            details=f"Successful connection for source \033[4m{source_to_test.name}\033[0m",
            start_time=start_time,
            full_path=None,
        )
        return JobResult(item=source_to_test, success=True, message=success_message)
    except Exception as e:
        failure_message = format_message_failure(
            details=f"Failed connection for source \033[4m{source_to_test.name}\033[0m",
            start_time=start_time,
            error_msg=f"Failed connection to host: {source_to_test.host}, port: {source_to_test.port}, username: {source_to_test.username}, database: {source_to_test.database}. {str(repr(e))}",
            full_path=None,
        )
        return JobResult(item=source_to_test, success=False, message=failure_message)


def job(source):
    return Job(
        item=source,
        source=source,
        action=action,
        source_to_test=source,
    )



import json
from visivo.models.project import Project
from visivo.models.trace_props.trace_props import TraceType
from importlib.resources import files


def generate_project_schema():
    schema = Project.model_json_schema(by_alias=False)
    schema["$schema"] = "https://json-schema.org/draft/2020-12/schema"
    schema_string = json.dumps(schema)
    schema_string = schema_string.replace("?P<ref_name>", "")
    schema_string = schema_string.replace("?P<column_name>", "")
    schema_string = schema_string.replace("?P<query_statement>", "")
    schema_string = schema_string.replace("?P<query_string>", "")

    return schema_string


def generate_trace_prop_schemas():
    """Generate schemas for all trace types"""
    schemas = {}
    for trace_type in TraceType:
        schema_path = files("visivo.schema").joinpath(f"{trace_type.value}.schema.json")
        with open(schema_path) as f:
            schemas[trace_type.value.capitalize()] = json.load(f)
    return schemas


def generate_schema():
    """Generate both project and trace prop schemas"""
    project_schema = json.loads(generate_project_schema())

    trace_schemas = generate_trace_prop_schemas()

    if "$defs" not in project_schema:
        project_schema["$defs"] = {}

    for trace_type, schema in trace_schemas.items():
        project_schema["$defs"][trace_type] = schema

    project_schema["$defs"]["Trace"]["properties"]["props"] = {
        "oneOf": [{"$ref": f"#/$defs/{trace_type}"} for trace_type in trace_schemas.keys()]
    }
    layout_schema = json.loads(files("visivo.schema").joinpath("layout.schema.json").read_text())
    project_schema["$defs"]["Layout"] = layout_schema
    project_schema["$defs"]["Trace"]["properties"]["layout"] = {"$ref": "#/$defs/Layout"}

    # Move nested $defs to top level
    defs_to_add = {}
    for def_name, def_schema in project_schema["$defs"].items():
        if "$defs" in def_schema:
            # Add each nested def to top level if it doesn't exist
            for nested_def_name, nested_def_schema in def_schema["$defs"].items():
                if (
                    nested_def_name not in project_schema["$defs"]
                    and nested_def_name not in defs_to_add
                ):
                    defs_to_add[nested_def_name] = nested_def_schema
            # Remove the nested $defs
            del def_schema["$defs"]

    # Add all collected defs to top level
    project_schema["$defs"].update(defs_to_add)
    return json.dumps(project_schema)



from visivo.logger.logger import Logger
from visivo.models.dag import all_descendants_of_type
from visivo.models.models.local_merge_model import LocalMergeModel
from visivo.models.models.model import Model
from visivo.models.models.csv_script_model import CsvScriptModel
from visivo.models.project import Project
from visivo.models.sources.source import Source
from visivo.models.trace import Trace
from visivo.query.aggregator import Aggregator
from visivo.jobs.job import (
    Job,
    JobResult,
    format_message_failure,
    format_message_success,
    start_message,
)
from time import time
from visivo.query.trace_tokenizer import TraceTokenizer
from visivo.query.query_string_factory import QueryStringFactory


def action(trace, dag, output_dir):
    Logger.instance().info(start_message("Trace", trace))
    model = all_descendants_of_type(type=Model, dag=dag, from_node=trace)[0]

    if isinstance(model, CsvScriptModel):
        source = model.get_duckdb_source(output_dir=output_dir)
    elif isinstance(model, LocalMergeModel):
        source = model.get_duckdb_source(output_dir=output_dir, dag=dag)
    else:
        source = all_descendants_of_type(type=Source, dag=dag, from_node=model)[0]

    trace_directory = f"{output_dir}/traces/{trace.name}"
    query_string = _get_query_string(trace, dag, output_dir)
    try:
        start_time = time()
        data_frame = source.read_sql(query_string)
        success_message = format_message_success(
            details=f"Updated data for trace \033[4m{trace.name}\033[0m",
            start_time=start_time,
            full_path=None,
        )
        Aggregator.aggregate_data_frame(data_frame=data_frame, trace_dir=trace_directory)
        return JobResult(item=trace, success=True, message=success_message)
    except Exception as e:
        if hasattr(e, "message"):
            message = e.message
        else:
            message = repr(e)
        failure_message = format_message_failure(
            details=f"Failed query for trace \033[4m{trace.name}\033[0m",
            start_time=start_time,
            full_path=None,
            error_msg=message,
        )
        return JobResult(item=trace, success=False, message=failure_message)


def _get_query_string(trace, dag, output_dir):
    model = all_descendants_of_type(type=Model, dag=dag, from_node=trace)[0]
    if isinstance(model, CsvScriptModel):
        source = model.get_duckdb_source(output_dir=output_dir)
    elif isinstance(model, LocalMergeModel):
        source = model.get_duckdb_source(output_dir=output_dir, dag=dag)
    else:
        source = all_descendants_of_type(type=Source, dag=dag, from_node=model)[0]
    tokenized_trace = TraceTokenizer(trace=trace, model=model, source=source).tokenize()
    return QueryStringFactory(tokenized_trace=tokenized_trace).build()


def _get_source(trace, dag, output_dir):
    sources = all_descendants_of_type(type=Source, dag=dag, from_node=trace)
    if len(sources) == 1:
        return sources[0]

    model = all_descendants_of_type(type=Model, dag=dag, from_node=trace)[0]
    if isinstance(model, CsvScriptModel):
        return model.get_duckdb_source(output_dir)
    elif isinstance(model, LocalMergeModel):
        return model.get_duckdb_source(output_dir, dag)
    else:
        return model.source


def job(dag, output_dir: str, trace: Trace):
    source = _get_source(trace, dag, output_dir)
    return Job(
        item=trace,
        source=source,
        action=action,
        trace=trace,
        dag=dag,
        output_dir=output_dir,
    )



import click
from visivo.commands.options import (
    working_dir,
    output_dir,
    dbt_profile,
    dbt_target,
)


@click.command()
@working_dir
@output_dir
@dbt_profile
@dbt_target
def dbt(working_dir, output_dir, dbt_profile, dbt_target):
    """
    Refreshes the dbt objects in the for use in Visivo.
    """
    from visivo.logger.logger import Logger

    Logger.instance().debug("Refreshing dbt models and sources.")

    from visivo.commands.dbt_phase import dbt_phase

    dbt_phase(
        working_dir=working_dir,
        output_dir=output_dir,
        dbt_profile=dbt_profile,
        dbt_target=dbt_target,
    )
    Logger.instance().success("Refreshed dbt models and sources.")



from typing import List
from pydantic import Field
from visivo.models.base.base_model import generate_ref_field
from visivo.models.base.parent_model import ParentModel
from visivo.models.dag import all_descendants_of_type
from visivo.models.models.csv_script_model import CsvScriptModel
from visivo.models.models.model import Model
from visivo.models.sources.duckdb_source import DuckdbAttachment, DuckdbSource
import click
import os

from visivo.models.sources.source import Source


class LocalMergeModel(Model, ParentModel):
    """
    Local Merge Models are models that allow you to merge data from multiple other models locally.

    !!! note

        Any joining is done in a local DuckDB database. While more efficient than SQLite,
        it's still primarily designed for medium-sized datasets.

    !!! example {% raw %}

        === "Internal join External"

            Here is an example of merging two models that are defined in your project. One that is external and one that is internal.
            ``` yaml
            models:
              - name: local_merge
                models:
                    - ref(first_domain_model)
                    - ref(external_data_model)
                sql: SELECT * FROM first_domain_model.model AS fdm JOIN external_data_model.model AS edm ON fdm.external_id = edm.id
            ```

    {% endraw %}
    """

    sql: str = Field(
        description="The sql used to generate your base data",
    )
    models: List[generate_ref_field(Model)] = Field(
        description="A model object defined inline or a ref() to a model."
    )

    def get_duckdb_source(self, output_dir, dag) -> DuckdbSource:
        os.makedirs(f"{output_dir}/models", exist_ok=True)
        attach = list(
            map(
                lambda model: DuckdbAttachment(
                    schema_name=model.name,
                    source=self._get_duckdb_from_model(model, output_dir, dag),
                ),
                self._get_dereferenced_models(dag),
            )
        )

        return DuckdbSource(
            name=f"model_{self.name}_generated_source",
            database=f"{output_dir}/models/{self.name}.duckdb",
            type="duckdb",
            attach=attach,
        )

    def _insert_dependent_models_to_duckdb(self, output_dir, dag):
        for model in self._get_dereferenced_models(dag):
            if isinstance(model, CsvScriptModel):
                continue  # CsvScriptModels are attached from their existing duckdb file.
            elif isinstance(model, LocalMergeModel):
                continue  # LocalMergeModels are attached from their existing duckdb file.
            else:
                duckdb_source = self._get_duckdb_from_model(model, output_dir, dag)
                source = all_descendants_of_type(type=Source, dag=dag, from_node=model)[0]
                data_frame = source.read_sql(model.sql)
                with duckdb_source.connect() as connection:
                    connection.execute("DROP TABLE IF EXISTS model")
                    connection.execute("CREATE TABLE model AS SELECT * FROM data_frame")

    def insert_duckdb_data(self, output_dir, dag):
        try:
            self._insert_dependent_models_to_duckdb(output_dir, dag)
        except Exception as e:
            raise click.ClickException(
                f"Failed to insert dependent models to duckdb for model {self.name}. Error: {str(e)}"
            )

        duckdb_source = self.get_duckdb_source(output_dir=output_dir, dag=dag)
        with duckdb_source.connect() as connection:
            data_frame = connection.execute(self.sql).pl()
            connection.execute("DROP TABLE IF EXISTS model")
            connection.execute("CREATE TABLE model AS SELECT * FROM data_frame")

    def _get_duckdb_from_model(self, model, output_dir, dag) -> DuckdbSource:
        if isinstance(model, CsvScriptModel):
            return model.get_duckdb_source(output_dir=output_dir)
        elif isinstance(model, LocalMergeModel):
            return model.get_duckdb_source(output_dir=output_dir, dag=dag)
        else:
            os.makedirs(f"{output_dir}/models", exist_ok=True)
            return DuckdbSource(
                name=f"model_{model.name}_generated_source",
                database=f"{output_dir}/models/{model.name}.duckdb",
                type="duckdb",
            )

    def _get_dereferenced_models(self, dag):
        models = all_descendants_of_type(type=Model, dag=dag, from_node=self)
        return list(filter(lambda model: model is not self, models))

    def child_items(self):
        return self.models



import platform
import os


def get_editor_configs():
    """Get editor configurations based on operating system"""
    system = platform.system()
    is_windows = system == "Windows"
    is_mac = system == "Darwin"
    is_linux = system == "Linux"

    # Helper to generate Windows paths for Program Files
    def win_program_files_paths(app_path):
        return [
            os.path.join(os.environ.get("PROGRAMFILES", ""), app_path),
            os.path.join(os.environ.get("PROGRAMFILES(X86)", ""), app_path),
            os.path.join(os.environ.get("LOCALAPPDATA", ""), app_path),
        ]

    # Helper function to get VS Code command for macOS
    def get_vscode_mac_command():
        # Check for CLI command first
        if os.path.exists("/usr/local/bin/code"):
            return ["/usr/local/bin/code", "--wait"]
        # Check for app existence and use open command as fallback
        elif os.path.exists("/Applications/Visual Studio Code.app"):
            return ["open", "-a", "Visual Studio Code", "--wait-apps"]
        return None

    editors = [
        {
            "name": "Visual Studio Code",
            "id": "vscode",
            "paths": {
                "windows": win_program_files_paths(r"Microsoft VS Code\bin\code.cmd")
                + [r"C:\Users\%USERNAME%\AppData\Local\Programs\Microsoft VS Code\bin\code.cmd"],
                "mac": [
                    "/usr/local/bin/code",
                    "/Applications/Visual Studio Code.app",
                    os.path.expanduser("~/Applications/Visual Studio Code.app"),
                ],
                "linux": ["/usr/bin/code", "/usr/local/bin/code"],
            },
            "commands": {
                "windows": ["code", "--wait"],
                "mac": get_vscode_mac_command(),
                "linux": ["code", "--wait"],
            },
        },
        {
            "name": "Cursor",
            "id": "cursor",
            "paths": {
                "windows": win_program_files_paths(r"Cursor\cursor.exe"),
                "mac": ["/Applications/Cursor.app/Contents/MacOS/Cursor"],
                "linux": ["/usr/bin/cursor", "/usr/local/bin/cursor"],
            },
            "commands": {
                "windows": ["cursor", "--wait"],
                "mac": ["cursor", "--wait"],
                "linux": ["cursor", "--wait"],
            },
        },
        {
            "name": "Sublime Text",
            "id": "sublime",
            "paths": {
                "windows": win_program_files_paths(r"Sublime Text\subl.exe"),
                "mac": [
                    "/usr/local/bin/subl",
                    "/Applications/Sublime Text.app/Contents/SharedSupport/bin/subl",
                ],
                "linux": ["/usr/bin/subl", "/usr/local/bin/subl"],
            },
            "commands": {
                "windows": ["subl", "--wait"],
                "mac": ["subl", "--wait"],
                "linux": ["subl", "--wait"],
            },
        },
        {
            "name": "Notepad++",
            "id": "notepadpp",
            "paths": {
                "windows": win_program_files_paths(r"Notepad++\notepad++.exe"),
                "mac": [],  # Not available on Mac
                "linux": ["/usr/bin/notepad++", "/usr/local/bin/notepad++"],
            },
            "commands": {
                "windows": ["notepad++", "-multiInst", "-nosession"],
                "mac": None,
                "linux": ["notepad++", "-multiInst", "-nosession"],
            },
        },
        {
            "name": "Atom",
            "id": "atom",
            "paths": {
                "windows": win_program_files_paths(r"Atom\atom.exe"),
                "mac": [
                    "/usr/local/bin/atom",
                    "/Applications/Atom.app/Contents/Resources/app/atom.sh",
                    os.path.expanduser("~/Applications/Atom.app/Contents/Resources/app/atom.sh"),
                ],
                "linux": ["/usr/bin/atom", "/usr/local/bin/atom"],
            },
            "commands": {
                "windows": ["atom", "--wait"],
                "mac": ["atom", "--wait"],
                "linux": ["atom", "--wait"],
            },
        },
        {
            "name": "Neovim",
            "id": "nvim",
            "paths": {
                "windows": [r"C:\Program Files\Neovim\bin\nvim.exe"],
                "mac": ["/usr/local/bin/nvim", "/opt/homebrew/bin/nvim"],
                "linux": ["/usr/bin/nvim", "/usr/local/bin/nvim"],
            },
            "commands": {"windows": ["nvim"], "mac": ["nvim"], "linux": ["nvim"]},
        },
        {
            "name": "TextMate",
            "id": "textmate",
            "paths": {
                "windows": [],  # Not available on Windows
                "mac": [
                    "/usr/local/bin/mate",
                    "/Applications/TextMate.app/Contents/Resources/mate",
                ],
                "linux": [],  # Not available on Linux
            },
            "commands": {"windows": None, "mac": ["mate", "--wait"], "linux": None},
        },
        {
            "name": "JetBrains WebStorm",
            "id": "webstorm",
            "paths": {
                "windows": win_program_files_paths(r"JetBrains\WebStorm\bin\webstorm64.exe"),
                "mac": [
                    "/usr/local/bin/webstorm",
                    "/Applications/WebStorm.app/Contents/MacOS/webstorm",
                ],
                "linux": ["/usr/bin/webstorm", "/usr/local/bin/webstorm"],
            },
            "commands": {
                "windows": ["webstorm64", "--wait"],
                "mac": ["webstorm", "--wait"],
                "linux": ["webstorm", "--wait"],
            },
        },
        {
            "name": "Vim",
            "id": "vim",
            "paths": {
                "windows": [r"C:\Program Files (x86)\Vim\vim82\vim.exe"],
                "mac": ["/usr/bin/vim", "/usr/local/bin/vim"],
                "linux": ["/usr/bin/vim", "/usr/local/bin/vim"],
            },
            "commands": {"windows": ["vim"], "mac": ["vim"], "linux": ["vim"]},
        },
    ]

    # Add platform-specific editors
    if is_windows:
        editors.append(
            {
                "name": "Notepad",
                "id": "notepad",
                "paths": {"windows": [r"C:\Windows\System32\notepad.exe"], "mac": [], "linux": []},
                "commands": {"windows": ["notepad"], "mac": None, "linux": None},
            }
        )

    if is_mac:
        editors.append(
            {
                "name": "BBEdit",
                "id": "bbedit",
                "paths": {
                    "windows": [],
                    "mac": [
                        "/usr/local/bin/bbedit",
                        "/Applications/BBEdit.app/Contents/Helpers/bbedit_tool",
                    ],
                    "linux": [],
                },
                "commands": {"windows": None, "mac": ["bbedit", "--wait"], "linux": None},
            }
        )

    return editors, "windows" if is_windows else "mac" if is_mac else "linux"



import threading
import time
from flask import Flask, request, jsonify
from threading import Event

from visivo.logger.logger import Logger
from visivo.tokens.token_functions import validate_and_store_token
from visivo.tokens.web_utils import generate_success_html_response

FLASK_PORT = 5001
token_received_event = Event()

callback_server = Flask(__name__)


def shutdown_server():
    time.sleep(1)
    shutdown = request.environ.get("werkzeug.server.shutdown")
    if shutdown:
        shutdown()


@callback_server.route("/authorize-device-token", methods=["GET", "POST"])
def authorize_device_callback():
    token = request.args.get("token") if request.method == "GET" else None
    if not token:
        return jsonify({"error": "Token not provided"}), 400

    Logger.instance().success("Received token via callback: " + token)
    validate_and_store_token(token)
    token_received_event.set()

    base_url = callback_server.config.get("BASE_URL")
    html_content = generate_success_html_response(base_url, timeout=5)

    threading.Thread(target=shutdown_server).start()
    return html_content, 200


def run_flask_server(base_url, port=5001):
    """
    Runs the Flask server on localhost:<port>.
    debug=False, use_reloader=False to avoid double-start issues.
    """
    callback_server.config["BASE_URL"] = base_url
    callback_server.run(host="localhost", port=port, debug=False, use_reloader=False)



from visivo.utils import DIST_PATH
from visivo.logger.logger import Logger
import traceback


def dist_phase(output_dir, dist_dir):
    import os
    import json
    import shutil
    from glob import glob
    import datetime
    from visivo.utils import get_dashboards_dir

    Logger.instance().info("Creating distribution for project in folder...")

    os.makedirs(f"{dist_dir}/data", exist_ok=True)

    try:
        dashboards_dir = get_dashboards_dir(output_dir)
        if os.path.exists(dashboards_dir):
            dist_dashboards_dir = os.path.join(dist_dir, "data", "dashboards")
            shutil.copytree(dashboards_dir, dist_dashboards_dir, dirs_exist_ok=True)

        with open(f"{output_dir}/project.json", "r") as f:
            project_json = json.load(f)
        created_at = (datetime.datetime.now().isoformat(),)
        with open(f"{dist_dir}/data/project.json", "w") as f:
            f.write(
                json.dumps(
                    {
                        "id": "id",
                        "project_json": project_json,
                        "created_at": created_at,
                    }
                )
            )
        with open(f"{dist_dir}/_redirects", "w") as f:
            f.write("/*    /index.html   200")

        with open(f"{dist_dir}/data/error.json", "w") as f:
            f.write(json.dumps({}))
        with open(f"{dist_dir}/data/project_history.json", "w") as f:
            f.write(json.dumps([{"created_at": created_at, "id": "id"}]))

        trace_dirs = glob(f"{output_dir}/traces/*/", recursive=True)
        for trace_dir in trace_dirs:
            trace_name = os.path.basename(os.path.normpath(trace_dir))
            if os.path.exists(f"{output_dir}/traces/{trace_name}/data.json"):
                os.makedirs(f"{dist_dir}/data/{trace_name}", exist_ok=True)
                shutil.copyfile(
                    f"{output_dir}/traces/{trace_name}/data.json",
                    f"{dist_dir}/data/{trace_name}/data.json",
                )

        shutil.copytree(DIST_PATH, dist_dir, dirs_exist_ok=True)
    except Exception as e:
        Logger.instance().error(
            f"Error creating dist. Try running `visivo run` to ensure your project is up to date."
        )
        Logger.instance().error(f"Message: {e}, set STACKTRACE=true to see full error")
        if os.environ.get("STACKTRACE"):
            Logger.instance().error(f"{traceback.format_exc()}")
            raise e



from typing import Any, Optional
import ast
import operator as op
import os
import json

from visivo.models.test_run import TestRun
from visivo.utils import (
    combine_dict_properties,
    merge_dicts,
    nested_dict_from_dotted_keys,
)

SUPPORTED_NUMPY_FUNCTIONS = ["sum", "all", "mean", "any"]
TRACE_DATA_NODES = ["props", "columns"]


def evaluate_expression(
    expression: str,
    project: Any = None,
    output_dir: str = "",
    test_run: Optional[TestRun] = None,
) -> Any:
    import numpy

    operators = {
        ast.Add: op.add,
        ast.Sub: op.sub,
        ast.Mult: op.mul,
        ast.Div: op.truediv,
        ast.Pow: op.pow,
        ast.BitXor: op.xor,
        ast.USub: op.neg,
    }

    def any_test_failed():
        if test_run is None:
            return False
        else:
            return test_run.success

    def build_list_of_nodes(node, previous_nodes=[]):
        if isinstance(node, ast.Subscript) or isinstance(node, ast.Attribute):
            previous_nodes.append(node)
            return build_list_of_nodes(node.value, previous_nodes)
        elif isinstance(node, ast.Name):
            previous_nodes.append(node)
            return previous_nodes
        else:
            return previous_nodes

    def get_object_from_data(trace):
        data_file_path = os.path.join(output_dir, trace.name, "data.json")

        if not os.path.exists(data_file_path):
            raise FileNotFoundError(f"Data file not found: {data_file_path}")

        with open(data_file_path, "r") as file:
            data = json.load(file)
            combined = combine_dict_properties(data)
            return nested_dict_from_dotted_keys(combined)

    def get_object_from_node(current_object, index, nodes):
        node = nodes[index]
        last_node = index == len(nodes) - 1
        next_index = index + 1
        if isinstance(node, ast.Subscript):
            value = current_object[node.slice.value]
            if last_node:
                return value
            return get_object_from_node(value, next_index, nodes)
        elif isinstance(node, ast.Attribute):
            if current_object.__class__.__name__ == "Trace" and node.attr in TRACE_DATA_NODES:
                trace_data = get_object_from_data(current_object)
                current_object = merge_dicts(current_object.model_dump(), trace_data)
            if isinstance(current_object, dict):
                value = current_object[node.attr]
            else:
                value = getattr(current_object, node.attr)
            if last_node:
                return value
            return get_object_from_node(value, next_index, nodes)
        elif isinstance(node, ast.Name):
            return current_object
        else:
            raise ValueError(f"Unsupported node type: {node.__class__.__name__}")

    def eval_subscripts(node):
        nodes = build_list_of_nodes(node)
        nodes = nodes[::-1]  # Reverse the order of nodes
        if hasattr(nodes[0], "id") and nodes[0].id == "project":
            value = get_object_from_node(project, 1, nodes)
            return value
        else:
            raise ValueError(f"Unsupported node type: {nodes[0].id}")

    def handle_numpy(node):
        if hasattr(numpy, node.func.id):
            numpy_function = getattr(numpy, node.func.id)
            return numpy_function(*[evaluate_node(arg) for arg in node.args])
        else:
            raise ValueError(f"Unsupported numpy function: {node.func.id}")

    def evaluate_node(node):
        if isinstance(node, ast.BoolOp) and isinstance(node.op, ast.And):
            return all(evaluate_node(value) for value in node.values)
        if isinstance(node, ast.Constant):
            return node.n
        if isinstance(node, ast.List):
            return list(map(lambda a: evaluate_node(a), node.elts))
        elif isinstance(node, ast.BinOp):
            return operators[type(node.op)](evaluate_node(node.left), evaluate_node(node.right))
        elif isinstance(node, ast.UnaryOp):
            return operators[type(node.op)](evaluate_node(node.operand))
        elif isinstance(node, ast.BoolOp):
            return operators[type(node.op)](evaluate_node(node.operand))
        elif isinstance(node, ast.Call):
            if node.func.id == "any_test_failed":
                return any_test_failed()
            elif node.func.id in SUPPORTED_NUMPY_FUNCTIONS:
                return handle_numpy(node)
            else:
                raise ValueError(f"Unsupported function: {node.func.id}")
        elif isinstance(node, ast.Compare):
            left = evaluate_node(node.left)
            for op, right in zip(node.ops, node.comparators):
                if not isinstance(op, (ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE)):
                    raise TypeError(f"Unsupported comparison operator: {op.__class__.__name__}")
                right = evaluate_node(right)
                if not compare_op(op, left, right):
                    return False
                left = right
            return True
        elif isinstance(node, ast.Attribute):
            if isinstance(node.value, ast.Name):
                if node.value.id == "env":
                    return os.getenv(node.attr)
                elif node.value.id == "project":
                    return getattr(project, node.attr)
            return eval_subscripts(node)
        else:
            raise TypeError(f"Unsupported type {node.__class__}")

    def compare_op(op, left, right):
        return {
            ast.Eq: lambda: left == right,
            ast.NotEq: lambda: left != right,
            ast.Lt: lambda: left < right,
            ast.LtE: lambda: left <= right,
            ast.Gt: lambda: left > right,
            ast.GtE: lambda: left >= right,
        }[type(op)]()

    try:
        parsed = ast.parse(expression, mode="eval")

        return evaluate_node(parsed.body)
    except (
        SyntaxError,
        TypeError,
        KeyError,
        ValueError,
        NameError,
        AttributeError,
        IndexError,
    ) as e:
        raise ValueError(f"Invalid expression: {expression}: {e}") from e



from typing import Any

import re

from visivo.models.dag import all_descendants_with_name

NAME_REGEX = r"a-zA-Z0-9\s'\"\-_"
INLINE_REF_REGEX = rf"\${{\s*ref\(([{NAME_REGEX}]+?)\)[\.\d\w\[\]]*\s*}}"
INLINE_REF_PROPS_PATH_REGEX = rf"\${{\s*ref\([{NAME_REGEX}]+?\)([\.\d\w\[\]]*)\s*}}"
INLINE_PATH_REGEX = rf"\${{\s*([{NAME_REGEX}\.\[\]]+?)\s*}}"
CONTEXT_STRING_VALUE_REGEX = rf"\${{\s*([{NAME_REGEX}\.\[\]\)\()]+?)\s*}}"


class ContextString:
    """
    Represents a string that contains a reference to named object in the project model.
    Currently you can reference another object by using the ref() function.  Example:

    ${ ref(Name) }
    """

    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, ContextString):
            return re.findall(CONTEXT_STRING_VALUE_REGEX, self.value) == re.findall(
                CONTEXT_STRING_VALUE_REGEX, other.value
            )
        return False

    def __hash__(self):
        return hash("".join(re.findall(CONTEXT_STRING_VALUE_REGEX, self.value)))

    def get_reference(self) -> str:
        matches = re.findall(INLINE_REF_REGEX, self.value)
        if len(matches) == 0:
            return None
        else:
            return matches[0]

    def get_ref_props_path(self) -> str:
        matches = re.findall(INLINE_REF_PROPS_PATH_REGEX, self.value)
        if len(matches) == 0:
            return None
        else:
            return matches[0]

    def get_path(self) -> str:
        matches = re.findall(INLINE_PATH_REGEX, self.value)
        if len(matches) == 0:
            return None
        else:
            return matches[0]

    def get_item(self, dag: Any) -> Any:
        reference = self.get_reference()
        items = all_descendants_with_name(reference, dag)
        if len(items) == 0:
            raise ValueError(f"Invalid context string reference name: '{reference}'.")
        return items[0]

    @classmethod
    def is_context_string(cls, obj) -> bool:
        return isinstance(obj, ContextString) or (
            isinstance(obj, str) and ContextString(obj).get_reference()
        )

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type: Any, handler: Any):
        from pydantic_core import core_schema

        def validate_and_create(value: Any) -> "ContextString":
            if isinstance(value, cls):
                return value
            str_value = str(value)
            if not (str_value.startswith("${") and str_value.endswith("}")):
                raise ValueError("ContextString must start with '${' and end with '}'")
            return cls(str_value)

        return core_schema.no_info_after_validator_function(
            validate_and_create,
            core_schema.str_schema(pattern=r"^\$\{.*\}$"),
            serialization=core_schema.plain_serializer_function_ser_schema(str),
        )



from visivo.utils import SCHEMA_FILE
from visivo.logger.logger import Logger
from importlib import reload, import_module
from visivo.parsers.schema_generator import generate_schema

"""
Generates the project schema json file for the current version of the CLI. 

The generated file is shipped with visivo installs and is used in the Viewer to 
validate schema client side and to power flask app api endpoints. 
"""


def reload_and_generate_schema():
    # Store original environment state

    trace_module = import_module("visivo.models.trace")
    reload(trace_module)

    project_module = import_module("visivo.models.project")
    reload(project_module)

    schema_string = generate_schema()
    if "?P<" in schema_string:
        raise Exception(
            f"Schema json contains '?P<': this means that the schema is likely not JS compatible and needs a new find and replace like the ones above this line"
        )

    return schema_string


def write_schema_json() -> None:
    """
    Write the Visivo Project schema for the current CLI version to the output directory
    if it doesn't exist for the current verison of visivo.

    Args:
        output_dir: Directory to write schema file
    """

    try:
        Logger.instance().info(f"Generating schema file: {SCHEMA_FILE}...")

        schema = reload_and_generate_schema()
        with open(SCHEMA_FILE, "w") as f:
            f.write(schema)
        Logger.instance().info(f"Schema file written to: {SCHEMA_FILE}")
    except Exception as e:
        Logger.instance().error(f"Error writing schema file: {str(e)}")
        raise e


if __name__ == "__main__":
    write_schema_json()



import re
from typing import List, Tuple
from networkx import descendants, ancestors
from networkx.algorithms.traversal import depth_first_search


def create_dag_dict(dag):
    if not hasattr(dag, "get_root_nodes"):
        raise ValueError("DAG is not a ProjectDag")
    project = dag.get_root_nodes()[0]
    dag.remove_node(project)
    dag = dag.reverse()
    nodes = []
    edges = []

    for node in dag.nodes():

        node_data = {
            "id": str(id(node)),
            "name": node.name,
            "path": node.path,
            "type": type(node).__name__.lower(),
        }
        nodes.append(node_data)

    for edge in dag.edges():
        edge_data = {"source": str(id(edge[0])), "target": str(id(edge[1]))}
        edges.append(edge_data)

    return {"nodes": nodes, "edges": edges}


def all_descendants(dag, from_node=None, depth=None):
    if not from_node:
        return set(dag.nodes())

    if not depth:
        descendants_list = list(descendants(dag, from_node))
        descendants_list.append(from_node)
        return set(descendants_list)

    # depth_first_search.dfs_tree is slow, so it is only used when depth is not None
    return depth_first_search.dfs_tree(dag, from_node, depth_limit=depth)


def all_descendants_of_type(type, dag, from_node=None, depth=None):
    if not depth:
        depth = len(dag)

    def find_type(item):
        return isinstance(item, type)

    return list(
        filter(
            find_type,
            all_descendants(dag=dag, from_node=from_node, depth=depth),
        )
    )


def all_descendants_with_name(name: str, dag, from_node=None):
    def find_name(item):
        return hasattr(item, "name") and item.name == name

    return list(filter(find_name, all_descendants(dag=dag, from_node=from_node)))


def all_descendants_with_path_match(path: str, dag, from_node=None):
    def path_match(item):
        return hasattr(item, "path") and item.path and item.path in path

    return list(filter(path_match, all_descendants(dag=dag, from_node=from_node)))


def family_tree_contains_named_node(item, name: str, dag):
    d = descendants(dag, item)
    a = ancestors(dag, item)
    items = d.union(a)
    return any(item.name == name for item in items)


def all_nodes_including_named_node_in_graph(name: str, dag):
    import click

    item = all_descendants_with_name(name=name, dag=dag)

    if len(item) == 1:
        item = item[0]
    else:
        raise click.ClickException(f"No item found with name: '{name}'.")

    d = descendants(dag, item)
    a = ancestors(dag, item)
    items = d.union(a)
    items.add(item)
    return items


def parse_filter_str(filter_str) -> List[Tuple[str, str, str]]:
    pattern = r"((?P<pre>\d*\+)?\s*(?P<name>[a-zA-Z0-9\s'\"\-_]+)\s*(?P<post>\+\d*)?)(,|$)"
    matches = re.finditer(pattern, filter_str)

    filters = []
    for match in matches:
        filters.append((match.group("pre"), match.group("name"), match.group("post")))
    return filters
